#include "CyanBus.hpp"

// C++ system level
#include <cstdio>     // sprintf
// #include <functional>

// ESP32 specific
#include "esp_log.h"
#include "driver/gpio.h"

// project specific
#include <Types.hpp>
#include "SiliconTorch/NVSExplorer.hpp"

// qthing stuff
#include <qthing>
#include "SiliconTorch/CyanBus.hpp"

// misc
#include <nlohmann/json.hpp>


using nlohmann::json;




namespace SiliconTorch {

  namespace Service {

    void CyanBus::init() {
      setIcon("🚌");
      setName("CyanBus");
      setNameSpace("CyanBus");
    }

    void CyanBus::start() {

      for (u8 idx = 0; idx < MAX_INTERFACES; idx++) {

        InterfaceCfg cfg = readInterfaceData(idx);

        if (!cfg.valid) break;

        ESP_LOGI(getName().c_str(), "Configuring: CyanBus{ tx[ %d ] rx[ %d ] de[ %d ] re[ %d ] baud[ %d ] interface[ cb%d ] }",
          cfg.tx,
          cfg.rx,
          cfg.de,
          cfg.re,
          cfg.baud,
          idx
        );

        ifaceCfgs.push_back(cfg);
        SiliconTorch::CyanBus::createCyanBusInterface(cfg.tx, cfg.rx, cfg.de, cfg.re, cfg.baud);

      }

    }


    CyanBus::InterfaceCfg CyanBus::readInterfaceData(u32 number) const {

      NVSExplorer::NVSExplorer& nvs = NVSExplorer::NVSExplorer::instance();

      InterfaceCfg cfg;

      char tmp[32];

      std::snprintf(tmp, sizeof(tmp), "cb%d.tx", number);
      cfg.tx = nvs.getUnsignedInt(getNameSpace(), std::string(tmp), cfg.tx);

      std::snprintf(tmp, sizeof(tmp), "cb%d.rx", number);
      cfg.rx = nvs.getUnsignedInt(getNameSpace(), std::string(tmp), cfg.rx);

      std::snprintf(tmp, sizeof(tmp), "cb%d.de", number);
      cfg.de = nvs.getUnsignedInt(getNameSpace(), std::string(tmp), cfg.de);

      std::snprintf(tmp, sizeof(tmp), "cb%d.re", number);
      cfg.re = nvs.getUnsignedInt(getNameSpace(), std::string(tmp), cfg.re);

      std::snprintf(tmp, sizeof(tmp), "cb%d.baud", number);
      cfg.baud = nvs.getUnsignedInt(getNameSpace(), std::string(tmp), cfg.baud);


      cfg.valid = cfg.tx < 0xFF
                && cfg.rx < 0xFF
                && cfg.de < 0xFF
                && cfg.re < 0xFF;

      return cfg;
    }


    json CyanBus::getConfigJSON() const {

      json out;

      u8 ch = 0;
      for (const auto& cfg : ifaceCfgs) {
        json channel;

        channel["tx"]   = cfg.tx;
        channel["rx"]   = cfg.rx;
        channel["de"]   = cfg.de;
        channel["re"]   = cfg.re;
        channel["baud"] = cfg.baud;

        out[ch++] = channel;
      }

      return out;
    }

  }
}


