#include "CyanBusInjector.hpp"

// C++ system level
#include <string>
#include <cstring>
#include <cstdio>     // snprintf
#include <cstdlib>    // strtol
#include <cinttypes>
#include <algorithm>
// #include <functional>

// ESP32 specific
// #include "esp_err.h"
// #include "esp_log.h"
// #include "driver/gpio.h"
// #include "driver/uart.h"
// #include "freertos/FreeRTOS.h"
// #include "freertos/queue.h"

// project specific
#include "CyanBusTX.hpp"
// #include "CyanBusCRC.hpp"

// qthing stuff
#include <qthing>
#include <qthing/mqtt_common.hpp>


// TODO: Needed…?
const char* const TAG = "CyanBusInjector";


namespace SiliconTorch {

  namespace CyanBusInjector {


    const std::string BreakHeader("[CyanBusBreak]");
    const std::string PacketHeader("[CyanBusPacket]");

    // TODO: Merge with other project tree and include this definition!
    const std::string CyanBusHeader("fxCyan");


    CyanBusInjector::CyanBusInjector(uint8_t tx, uint8_t rx, uint8_t de, uint8_t re, uint32_t baudRate, uint8_t uartChannel) : cbTX(tx, rx, de, re, baudRate, uartChannel) {

      auto packetHandler = [&](qthing::udpPacket packet) {

        bool _break = false;
        uint8_t headerLength = 0;

        // Check packet type…
        if (packet.payload[8] == 'B') {

          _break = true;
          headerLength = BreakHeader.length();

        } else if (packet.payload[8] == 'P') {

          _break = false;
          headerLength = PacketHeader.length();

        } else {
          ESP_LOGW(TAG, "PacktHandler got unsupported packet type! packet{ [8] = %c }", packet.payload[8]);
          return;
        }

        cbTX.write((uint8_t*)(packet.payload + headerLength), packet.length - headerLength, _break);


        // ⬇⬇ Handling BaudRate-packets ⬇⬇

        /*
        uint16_t packetLength = packet.length - headerLength;

        uint8_t minLength = CyanBusHeader.length()
                            + 2                     // length header
                            + 4                     // CRC header
                            + 6;                    // length of "fxBaud"

        if (packetLength > minLength) {

          char* _packet = (char*)(packet.payload + headerLength);

          std::string fxCyan("fxCyan");
          if (strncmp(_packet, fxCyan.c_str(), fxCyan.length()) != 0) return;  // this is not the beginning of a CyanBus packet


          _packet += fxCyan.length();

          uint16_t _length = (_packet[0] << 8) | packet[1];  // CyanBus packet's length header
          _packet += 2;

          std::string fxBaud("fxBaud");
          if (strncmp(_packet, fxBaud.c_str(), fxBaud.length()) != 0) return;  // this is not the beginning of a BaudRate packet

          uint16_t remainingPacketLength = packetLength - (_packet - packet.payload);  // remaining length from our pointer to the end of UDP packet

          bool crcOK = SiliconTorch::Impl::checkCRC(_packet, std::min(_length, remainingPacketLength));

          // bail out before we change the baud rate and send a packet which will get discarded by nodes
          if (!crcOK) return;

          // TODO: parse and change baud rate!

        }
        */

        // ⬆⬆ Handling BaudRate-packets ⬆⬆
      };

      qthing::addUDPPacketCallback(BreakHeader, packetHandler);
      qthing::addUDPPacketCallback(PacketHeader, packetHandler);


      /*std::function<void(const std::string&)>*/
      
      auto setBaud = [&](const std::string& message) {
        long int baud = strtol(message.c_str(), NULL, 0);
        this->setBaudRate(baud);
      };

      auto getBaud = [&](const std::string& ignored) {
        this->publishBaudRate();
      };

      qthing::add_message_callback(this->deviceTopic("baudRate/set"), setBaud);
      qthing::add_message_callback(this->serviceTopic("baudRate/set"), setBaud);

      qthing::add_message_callback(this->deviceTopic("baudRate/get"), getBaud);
      qthing::add_message_callback(this->serviceTopic("baudRate/get"), getBaud);

    }

    void CyanBusInjector::setBaudRate(uint32_t baud) {
      cbTX.setBaudRate(baud);
    }

    uint32_t CyanBusInjector::getBaudRate() {
      return cbTX.getBaudRate();
    }

    void CyanBusInjector::publishBaudRate() {
      char tmp[16];
      snprintf(tmp, sizeof(tmp), "%d", this->getBaudRate());
      qthing::publish_message(this->deviceTopic("baudRate"), tmp);
    }

    std::string CyanBusInjector::deviceTopic(const char *suffix) {
      return std::string(DEVICE_NAMESPACE + "CyanBus/") + std::string(suffix);
    }

    std::string CyanBusInjector::serviceTopic(const char *suffix) {
      return std::string("service/CyanBus/") + std::string(suffix);
    }


  }
}
